function mergeBuffer(...buffers) {
    const Cons = buffers[0].constructor;
    return buffers.reduce((pre, val) => {
        const merge = new Cons((pre.byteLength | 0) + (val.byteLength | 0));
        merge.set(pre, 0);
        merge.set(val, pre.byteLength | 0);
        return merge;
    }, new Cons());
}

function debounce(func, wait, context) {
    let timeout;
    return function fn(...args) {
        const later = function later() {
            timeout = null;
            func.apply(context, args);
        };
        clearTimeout(timeout);
        timeout = setTimeout(later, wait);
    };
}

export default class Dida {
    constructor(option = {}) {
        this.option = {
            ...Dida.option,
            ...option,
        };

        this.context = new (window.AudioContext || window.webkitAudioContext)();
        this.gainNode = this.context.createGain();
        this.gainNode.gain.value = this.option.volume;
        this.source = null;

        this.decoding = false;
        this.playing = false;
        this.loadLength = 0;
        this.loadByteSize = 0;
        this.audioDuration = 0;
        this.audioLength = 0;

        this.timestamps = [];
        this.audiobuffers = [];
        this.timestampTmp = [];
        this.decodeErrorBuffer = new Uint8Array();
        this.decodeWaitingBuffer = new Uint8Array();

        this.autoEndDebounce = debounce(() => {
            this.end();
        }, this.option.autoEndTime);
    }

    static get option() {
        return {
            volume: 0.7,
            cache: true,
            chunk: 64 * 1024,
            autoEnd: true,
            autoEndTime: 5000,
            onNext: t => t,
            onLoad: () => null,
            onStop: () => null,
            onPlay: () => null,
            onEnd: () => null,
            onDestroy: () => null,
            onDecodeDone: () => null,
            onDecodeError: () => null,
            onVolumeChange: () => null,
        };
    }

    get volume() {
        return this.gainNode.gain.value;
    }

    set volume(value) {
        if (this.volume !== value) {
            this.gainNode.gain.value = value;
            this.option.onVolumeChange(value);
        }
    }

    destroy() {
        this.stop();
        this.context = null;
        this.gainNode = null;
        this.source = null;
        this.timestamps = [];
        this.audiobuffers = [];
        this.timestampTmp = [];
        this.decodeErrorBuffer = new Uint8Array();
        this.decodeWaitingBuffer = new Uint8Array();
        this.option.onDestroy();
        return this;
    }

    load(uint8, timestamp) {
        this.decoding = true;
        this.loadLength += 1;
        this.loadByteSize += uint8.byteLength;
        this.option.onLoad(uint8, timestamp);
        if (this.decodeWaitingBuffer.byteLength >= this.option.chunk) {
            this.timestamps.push(this.timestampTmp[0]);
            this.timestampTmp = [];
            const { buffer } = mergeBuffer(this.decodeErrorBuffer, this.decodeWaitingBuffer);
            this.decodeWaitingBuffer = new Uint8Array();
            this.context.decodeAudioData(
                buffer,
                audiobuffer => {
                    this.audioDuration += audiobuffer.duration;
                    this.audioLength += audiobuffer.length;
                    this.audiobuffers.push(audiobuffer);
                    this.decodeErrorBuffer = new Uint8Array();
                    this.option.onDecodeDone(audiobuffer);
                },
                error => {
                    this.decodeErrorBuffer = mergeBuffer(this.decodeErrorBuffer, this.decodeWaitingBuffer);
                    this.option.onDecodeError(error);
                },
            );
        } else {
            this.timestampTmp.push(timestamp);
            this.decodeWaitingBuffer = mergeBuffer(this.decodeWaitingBuffer, uint8);
        }
        if (this.option.autoEnd) {
            this.autoEndDebounce();
        }
        return this;
    }

    end() {
        if (this.decodeWaitingBuffer.length) {
            this.timestamps.push(this.timestampTmp[0]);
            this.timestampTmp = [];
            const { buffer } = this.decodeWaitingBuffer;
            this.decodeWaitingBuffer = new Uint8Array();
            this.decodeErrorBuffer = new Uint8Array();
            this.context.decodeAudioData(buffer, audiobuffer => {
                this.audioDuration += audiobuffer.duration;
                this.audioLength += audiobuffer.length;
                this.audiobuffers.push(audiobuffer);
                this.decoding = false;
                this.option.onEnd();
            });
        }
        return this;
    }

    play(startTime = 0) {
        if (this.source) {
            this.source.onended = null;
            this.source.stop();
            this.source = null;
        }
        this.playing = true;
        const index = this.timestamps.findIndex((timestamp, i) => {
            const audiobuffer = this.audiobuffers[i];
            return audiobuffer && timestamp + audiobuffer.duration * 1000 >= startTime;
        });
        const timestamp = this.timestamps[index];
        const audiobuffer = this.audiobuffers[index];
        if (timestamp === undefined || audiobuffer === undefined) return this.stop(index, timestamp);
        const offset = Math.max(0, (startTime - timestamp) / 1000);
        this.source = this.context.createBufferSource();
        this.source.connect(this.gainNode);
        this.gainNode.connect(this.context.destination);
        this.source.buffer = audiobuffer;
        this.option.onPlay(audiobuffer, startTime, offset);
        this.source.start(0, offset);
        this.source.onended = () => {
            const nextTimestamp = this.timestamps[index + 1];
            const nextAudiobuffer = this.audiobuffers[index + 1];
            if (nextTimestamp !== undefined && nextAudiobuffer !== undefined) {
                this.play(this.option.onNext(nextTimestamp));
                if (!this.option.cache) {
                    this.audiobuffers.splice(0, index + 1);
                    this.timestamps.splice(0, index + 1);
                }
            } else {
                this.stop(index, timestamp);
            }
        };
        return this;
    }

    stop(index, timestamp) {
        this.playing = false;
        if (this.source) {
            this.source.onended = null;
            this.source.stop();
            this.source = null;
        }
        this.option.onStop(index, timestamp);
        return this;
    }
}
